#    Copyright 2021 Qruise project
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qcalibrateremote._grpc import combine_key_and_secret, split_key_and_secret
from cuid import cuid
from icecream import ic

def test_combine_and_split():
    key = cuid()
    secret = cuid()
    ic(key, secret)
    
    keyAndSecret = combine_key_and_secret(client_key=key, client_secret=secret)
    ic(keyAndSecret)
    (k,s) = split_key_and_secret(keyAndSecret)
    assert key == k
    assert secret == s



#    Copyright 2021 Qruise project
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from base64 import b64decode, b64encode

GRPC_AUTH_HEADER = 'rpc-auth-header'

_ASCII = "ascii"

def split_key_and_secret(keyAndSecret):
    (client_key, client_secret) = b64decode(str.encode(keyAndSecret, _ASCII)).decode(_ASCII).split(':', 2)
    return client_key,client_secret

def combine_key_and_secret(client_key, client_secret):
    return b64encode(f"{client_key}:{client_secret}".encode(_ASCII)).decode(_ASCII)



#    Copyright 2021 Qruise project
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys
from typing import Optional, Union

from ._endpoint import Endpoint
from ._observer import OptimizationObserver


def _ipython_info():
    ip = None
    if 'ipykernel' in sys.modules:
        ip = 'notebook'
    elif 'IPython' in sys.modules:
        ip = 'terminal'
    return ip


def _open_browser(url: str):
    if _ipython_info():
        from IPython.display import Javascript, display
        display(Javascript('window.open("{url}");'.format(url=url)))


class BrowserPresenter(OptimizationObserver):
    def __init__(
        self,
        webfrontend_url: str,
    ) -> None:
        self._webfrontend_url = webfrontend_url

    def on_start(self, run_id: str, *args, **kwargs):
        """Opens browser window showing current run's details, if running a notebook in a browser

        Parameters
        ----------
        run_id : [type]
            [description]
        """
        from urllib.parse import urljoin
        path = f"run/{run_id}"
        _open_browser(urljoin(self._webfrontend_url, path))

    @staticmethod
    def from_endpoint(endpoint: Endpoint) -> Optional[OptimizationObserver]:
        """Creates Browser Presenter for a given endpoint for development and deployed environment
        expect host to be something grpc.* or localhost

        Parameters
        ----------
        endpoint : Endpoint
            remote services endpoint to start with

        Returns
        -------
        Optional[OptimizationObserver]
            instance or None
        """
        host = endpoint.host
        if host.startswith("grpc"):
            www_host = host.replace("grpc", "www")
            return BrowserPresenter(webfrontend_url=f"https://{www_host}:{endpoint.port}")
        elif host == "localhost" and endpoint.port == 9111:
            # development environment
            return BrowserPresenter(webfrontend_url="http://localhost:3000")
        else:
            return None



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import qcalibrate_pb2 as qcalibrate__pb2


class QCalibrateStub(object):
    """The qcalibrate remote service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddExperiment = channel.unary_unary(
                '/qcalibrateremote._grpc.QCalibrate/AddExperiment',
                request_serializer=qcalibrate__pb2.AddExperimentRequest.SerializeToString,
                response_deserializer=qcalibrate__pb2.AddExperimentResult.FromString,
                )
        self.Run = channel.stream_stream(
                '/qcalibrateremote._grpc.QCalibrate/Run',
                request_serializer=qcalibrate__pb2.Request.SerializeToString,
                response_deserializer=qcalibrate__pb2.Response.FromString,
                )


class QCalibrateServicer(object):
    """The qcalibrate remote service definition.
    """

    def AddExperiment(self, request, context):
        """runs optimization session
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Run(self, request_iterator, context):
        """runs optimization session
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_QCalibrateServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AddExperiment': grpc.unary_unary_rpc_method_handler(
                    servicer.AddExperiment,
                    request_deserializer=qcalibrate__pb2.AddExperimentRequest.FromString,
                    response_serializer=qcalibrate__pb2.AddExperimentResult.SerializeToString,
            ),
            'Run': grpc.stream_stream_rpc_method_handler(
                    servicer.Run,
                    request_deserializer=qcalibrate__pb2.Request.FromString,
                    response_serializer=qcalibrate__pb2.Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qcalibrateremote._grpc.QCalibrate', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class QCalibrate(object):
    """The qcalibrate remote service definition.
    """

    @staticmethod
    def AddExperiment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qcalibrateremote._grpc.QCalibrate/AddExperiment',
            qcalibrate__pb2.AddExperimentRequest.SerializeToString,
            qcalibrate__pb2.AddExperimentResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Run(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/qcalibrateremote._grpc.QCalibrate/Run',
            qcalibrate__pb2.Request.SerializeToString,
            qcalibrate__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#    Copyright 2021 Qruise project
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dataclasses import dataclass
from typing import Dict, List

from ._fom import Pulse


@dataclass
class IterationResult:
    iteration: int
    parameters: Dict[str, float]
    pulses: Dict[str, Pulse]
    figure_of_merit: float
    data: str


@dataclass
class PulseMeta:
    """Pulse meta parameter"""
    index: int
    """ordinal of a pulse in pulse meta list"""
    name: str
    """name of the pulse"""
    configuration: str
    """JSON serialized pulse optimization meta-parameters like basis, sampling, initial guess and boundaries"""


@dataclass
class ParameterMeta:
    """Parameter optimization meta"""
    index: int
    """ordinal of a parameter in a parameter list"""
    name: int
    """name of a parameter"""
    configuration: str
    """parameter optimization configuration, like boundaries and initial guess"""


@dataclass
class RunMetaParameters:
    """run attributes"""
    run_id: str
    """database ID"""
    configuration: str
    """JSON serialized optimization parameters"""
    parameter_metas: List[ParameterMeta]
    """List of optimized parameters"""
    pulse_metas: List[PulseMeta]
    """List of optimized pulses"""


class OptimizationObserver:
    """Base optimization observer
        don't do anything
    """

    def on_start(
            self,
            run_id: str,
            configuration: str,
            parameter_metas: List[ParameterMeta],
            pulse_metas: List[PulseMeta],
            **kwargs):
        """Invoked after the run entity is created

        Parameters
        ----------
        run_id : str
            The run entity id, can be used to navigate user to the run details page
        configuration : str
            JSON serialized configuration
        parameter_metas : List[ParameterMeta]
            list of metadata (like range and initial value) of optimized parameters
        pulse_metas : List[PulseMeta]
            list of metadata (like base, range and initial value) of optimized pulses
        """
        ...

    def on_iteration(
            self,
            iteration: int,
            parameters: Dict[str, float],
            pulses: Dict[str, Pulse],
            figure_of_merit: float,
            data: str,
            **kwargs):
        """invoked after the figure of merit is evaluated by the client

        Parameters
        ----------
        iteration : int
            the number of iteration (starting with 0)
        parameters : Dict[str, float]
            optimization parametes to evaluate, the key matches parameter name as defined in experiment configuration
        pulses : Dict[str, Pulse]
            pulse parametes to evaluate, the key matches parameter name as defined in experiment configuration
        figure_of_merit : float
            the evaluated figure of merit value
        data : str
            arbitrary data, returned by the figure of merit evaluator
        """
        ...

    def on_end(self):
        """Called as a last before the optimization is finished, can be used to free resources like network connection or mutexes
        """
        ...

    def on_error(self, exception: Exception):
        """Invoked if optimization is interrupted due to an exception

        Parameters
        ----------
        exception : Exception
            The exception caused the optimization run abortion
        """
        ...


class OptimizationResultCollector(OptimizationObserver):
    """Stores intermediate iteration results
    """

    def __init__(self) -> None:
        super().__init__()
        self._parameters: RunMetaParameters = None
        self._iterations: List[IterationResult] = []
        self._exception: Exception = None
        self._top: List[IterationResult] = []

    def on_start(self, run_id, configuration: str, parameter_metas, pulse_metas, **kwargs):
        self._parameters = RunMetaParameters(
            run_id=run_id,
            configuration=configuration,
            parameter_metas=parameter_metas,
            pulse_metas=pulse_metas)

    def on_iteration(self,
                     iteration: int,
                     parameters: Dict[str, float],
                     pulses: Dict[str, Pulse],
                     figure_of_merit: float,
                     data: str,
                     **kwargs):
        iteration_result = IterationResult(
            iteration=iteration,
            parameters=parameters,
            pulses=pulses,
            figure_of_merit=figure_of_merit,
            data=data)

        self._iterations.append(iteration_result)

        if len(self._top) == 0:
            self._top.append(iteration_result)
        else:
            # TODO: implement parametrically multiple top
            if self._top[0].figure_of_merit > iteration_result.figure_of_merit:
                self._top[0] = iteration_result

    def on_error(self, exception: Exception):
        self._exception = exception

    @property
    def parameters(self) -> RunMetaParameters:
        """ gets optimization run parameters

        Returns
        -------
        RunMetaParameters
            optimization run parameters
        """
        return self._parameters

    @property
    def iterations(self) -> List[IterationResult]:
        """ gets iteration results

        Returns
        -------
        List[IterationResult]
            iteration results, last iteration last
        """
        return self._iterations

    @property
    def exception(self) -> Exception:
        """ gets         

        Returns
        -------
        Exception
            exception in case of error
        """
        return self.exception

    @property
    def top(self) -> List[IterationResult]:
        """ gets best iterations (currently just the single one)

        Returns
        -------
        List[IterationResult]
            iteration(s) with  the lowest figure of merit value, the best (lowest figure of merit) first
        """
        return self._top



#    Copyright 2021 Qruise project
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from ._client import QOptimizerClient, create_optimizer_client
from ._endpoint import Endpoint, EndpointSecureKey
from ._fom import (EvaluateFigureOfMerit, EvaluateFigureOfMeritFactory,
                  FigureOfMerit, Pulse)
from ._observer import (IterationResult, OptimizationObserver,
                       RunMetaParameters, ParameterMeta, PulseMeta, OptimizationResultCollector)

from ._browser import BrowserPresenter

__all__ = [
    "create_optimizer_client",
    "BrowserPresenter",
    "Endpoint",
    "EndpointSecureKey",
    "EvaluateFigureOfMerit",
    "EvaluateFigureOfMeritFactory",
    "FigureOfMerit",
    "IterationResult",
    "OptimizationObserver", 
    "RunMetaParameters",
    "ParameterMeta",
    "Pulse",
    "PulseMeta",
    "QOptimizerClient",
    "OptimizationResultCollector",
]
